require 'rexml/document'
require 'facter'
require_relative './core'
if Facter.value(:kernel) == 'windows'
  require 'win32/registry'
end

module PuppetX
  module EnterpriseModules
    module OraInstall
      module Facts
        include ::PuppetX::EnterpriseModules::OraInstall::GetHomes

        def os
          Facter.value(:kernel)
        end

        def get_database_user(home)
          databaseUser = Facter.value('override_database_user')
          if databaseUser.nil?
            # puts "database user is oracle"
            if File.directory?("#{home}/bin")
              Facter::Util::Resolution.exec("/bin/stat --format=%U #{home}/bin/oracle")
            else
              'oracle'
            end
          else
            # puts "database user is " + databaseUser
            return databaseUser
          end
        end

        def get_su_command
          if os == 'Linux'
            return 'su -l '
          elsif os == 'SunOS'
            return 'su - '
          end
          'su -l '
        end

        def get_ora_inv_path
          if 'Linux' == os
            return '/etc'
          elsif 'SunOS' == os
            return '/var/opt/oracle'
          end
          '/etc'
        end

        def get_opatch_version(home)
          version = lsinventory(home).scan(/OPatch [V|v]ersion\s*:\s*(.*)/).flatten.first
          Puppet.debug "ora_install opatch version #{home}/OPatch/opatch: #{version}"
          (version.nil?) ? 'Invalid' : version
        end

        def get_orainst_loc
          if os == 'windows'
            begin
              str = Win32::Registry::HKEY_LOCAL_MACHINE.open('SOFTWARE\\Oracle').read('inst_loc')[1]
              return str
            rescue
              return ''
            end
          else
            if File.exist?(get_ora_inv_path + '/oraInst.loc')
              str = ''
              output = File.read(get_ora_inv_path + '/oraInst.loc')
              output.split(/\r?\n/).each do |item|
                if item.match(/^inventory_loc/)
                  str = item[14, 50]
                end
              end
              return str
            else
              return ''
            end
          end
        end

        def lsinventory(home)
          @lsinventory ||= {}
          if os == 'windows'
            @lsinventory[home] ||= `#{home}/OPatch/opatch lsinventory`
          else
            if supports_nolog?(home)
              @lsinventory[home] ||= Facter::Util::Resolution.exec("su #{user_for_file(home)} -c 'ORACLE_HOME=#{home} #{home}/OPatch/opatch lsinventory -customLogDir /tmp'")
              # We remove the created directory here, because it sometimes creates a dir with an other owner, which causes troubles for the next oracle home
              Puppet.debug `rm -rf /tmp/opatch`
              @lsinventory[home]
            else
              @lsinventory[home] ||= Facter::Util::Resolution.exec("su #{user_for_file(home)} -c 'ORACLE_HOME=#{home} #{home}/OPatch/opatch lsinventory'")
            end
          end
        end

        def supports_nolog?(home)
          @lsinventory_help ||= {}
          @lsinventory_help[home] ||= Facter::Util::Resolution.exec("su #{user_for_file(home)} -c 'ORACLE_HOME=#{home} #{home}/OPatch/opatch lsinventory -help'")
          @lsinventory_help[home] =~ /customLogDir/
        end

        def get_homes(path)
          unless path.nil?
            if File.exist?(path + '/ContentsXML/inventory.xml')
              file = File.read(path + '/ContentsXML/inventory.xml')
              doc = REXML::Document.new(file)
              homes = []
              doc.elements.each('/INVENTORY/HOME_LIST/HOME') do |element|
                #
                # Skip all entryes that have subelements.
                #
                next if element.elements[1] && element.elements[1].name == 'DEPHOMELIST'
                str = element.attributes['LOC']
                name = element.attributes['NAME']
                if element.attributes['REMOVED'] == 'T'
                  Puppet.debug "Home #{str} marked as removed in Oracle inventory; skipping for ora_install_homes fact."
                  next
                end
                unless str.nil?
                  if str.include? 'plugins'
                    # skip EM agent
                  elsif str.include? 'jdk'
                    # skip EM agent
                  elsif str.include? 'OraPlaceHolderDummyHome'
                    # skip EM agent
                  else
                    homes << str
                  end
                end
              end
              return homes
            else
              return {}
            end
          else
            return {}
          end
        end

        def get_home_name(home)
          inventory_loc = get_orainst_loc
          if File.exist?(inventory_loc + '/ContentsXML/inventory.xml')
            file = File.read(inventory_loc + '/ContentsXML/inventory.xml')
            doc = REXML::Document.new(file)
            doc.elements.each('/INVENTORY/HOME_LIST/HOME') do |element|
              if element.attributes['LOC'] == home
                return element.attributes['NAME']
              end
            end
          else
            return 'Invalid'
          end
        end

        def get_version(home)
          version = opatch_installed?(home) ? lsinventory(home).scan(/^(?:Oracle Database|Oracle Client|Oracle Grid Infrastructure|Oracle GoldenGate)\s.+?\s+(.*)$/).flatten.first : 'Invalid'
          (version.nil?) ? 'Invalid' : version
        end

        def get_installed_patches(oracle_home)
          if File.directory?(oracle_home)
            unless os == 'windows'
              ora_user = user_for_file(oracle_home)
            end
            installed_patches = patches_in_home(oracle_home, ora_user).flatten.sort
          else
            installed_patches = ['Invalid']
          end
          cleanup_opatch_residue(oracle_home) if Facter.value(:kernel) == 'Linux'
          installed_patches
        end

        def patches_in_home(oracle_product_home_dir, os_user)
          if opatch_installed?(oracle_product_home_dir)
            lsinventory(oracle_product_home_dir).scan(/Patch\s.(\d+)\s.*:\sapplied on/).flatten
          else
            #
            # When Opatch is not installed, we don't report any patches
            #
            ['Invalid']
          end
        end

        def opatch_installed?(oracle_home)
          File.exist?("#{oracle_home}/OPatch/opatch")
        end

        def user_for_file(home)
          #
          # To run opatch, we need the owner of the grid and oracle directory. The mechanism
          # we use to get this user, is to fetch the owner of the bin directory in the grid or
          # oracle home. Sometimes however, these directories also contain files with root as owner.
          # since root can never be the user of the grid or oracle processen, we skip those.
          #
          if os == 'windows'
            # On Windows all files and dirs have uid=0, use 'oracle' for now
            # TODO: find a way to determine the real owner on Windows
            'oracle'
          else
            a_file = first_non_root_file(home)
            uid = File.stat("#{home}/#{a_file}").uid
            Etc.getpwuid(uid).name
          end
        end

        def first_non_root_file(dir)
          Dir.new(dir).each { |file| break file if File.stat("#{dir}/#{file}").uid.nonzero? }
        end

        def cleanup_opatch_residue(home)
          # Some older versions of Opatch leave files and directories arround
          # When running opatch often, this causes some issues. This is to make
          # sure all files created by running Opatch as part of the fact, are remopved
          basename = Time.now.strftime("%Y-%m-%d_%H-%M-*%p")
          Puppet.debug `rm -rf /tmp/OraInstall*`
          Puppet.debug `rm -rf /tmp/opatch`
          FileUtils.rm_f("#{home}/cfgtoollogs/opatch/lsinv/lsinventory#{basename}.txt")
        end

        def get_defined_sids(home)
          if os == 'windows'
            home_name = get_home_name(home)
            regkey = Win32::Registry::HKEY_LOCAL_MACHINE.open("SOFTWARE\\Oracle\\KEY_#{home_name}")
            regkey.select { |k,_v1, v| k.scan(/ORA_(.*)_AUTOSTART/) != [] }.flatten.collect {|k| k.to_s.scan(/ORA_(.*)_AUTOSTART/)}.flatten.uniq
          else
            quoted_home = Regexp.quote(home)
            oratab_content.scan(/^([+-_A-Za-z].*):#{quoted_home}:[YN].*$/).flatten.uniq
          end
        end

        def get_running_processes(home)
          if os == 'Linux'
            {
              'sids'      => get_running_sids(home),
              'listeners' => get_running_listeners(home)
            }
          else
            {
              'sids'      => 'n/a',
              'listeners' => 'n/a'
            }
          end
        end

        def get_running_sids(home)
          command = %Q(ps aux | egrep "(ora|asm)_pmon" | awk '{gsub("(ora|asm)_pmon_", ""); print $2, $NF}' |
                       while read pid sid; do
                           printf "%-9s,%-80s\n" $sid  $(ls -l /proc/$pid/exe | awk '{ print $NF }' | sed 's/\\/bin\\/oracle$//' | sort | uniq)
                       done)
          raw_info = Facter::Util::Resolution.exec(command)
          if $?.success?
            return {} if raw_info.empty?
            databases = raw_info.split("\n").select{|e| e.include?(home)}.collect {|e| e.split.first}.sort
            db_hash = databases.collect do |db|
              if db =~ /ASM/
                db_mode_role = ['n/a','n/a']
              else
                db_mode_role = get_mode_role(db, home).split
              end
              {db => { 'open_mode'     => db_mode_role[0],
                       'database_role' => db_mode_role[1] }}
            end
            Hash[*db_hash.collect {|h| h.to_a}.flatten]
          else
            puts "Fetching running SIDS, needs access to /proc/pid/exe file system."
            puts "Probably using docker in non-privileged mode."
            puts "Run docker in privileged mode to fix the issue."
            nil
          end
        end

        def get_running_listeners(home)
          command = %Q(ps -eo pid,cmd | grep "#{home}/bin/tns[l]snr" | sed 's/-inherit//' | sed 's/-no_crs_notify//' |
                       while read pid command listener; do
                           printf "%-9s %-80s\n" $listener  $(ls -l /proc/$pid/exe | awk '{ print $NF }' | sort | uniq)
                       done)
          raw_info = Facter::Util::Resolution.exec(command)
          if $?.success?
            return [] if raw_info.empty?
            raw_info.split("\n").select{|e| e.include?(home)}.collect {|e| e.split.first}.sort
          else
            puts "Fetching running listeners, needs access to /proc/pid/exe file system."
            puts "Probably using docker in non-privileged mode."
            puts "Run docker in privileged mode to fix the issue."
            nil
          end
        end

        def get_mode_role(sid, home)
          command     = %Q(sqlplus -S /nolog << EOF
            connect / as sysdba
            set heading off
            select replace(open_mode, chr(32), chr(95)) as open_mode
                 , replace(database_role, chr(32), chr(95)) as database_role
              from v\\$database;
EOF)
          `su #{user_for_file(home)} -c 'export ORACLE_HOME="#{home}"; \
                                         export PATH="#{home}/bin:$PATH"; \
                                         export ORACLE_SID="#{sid}"; \
                                         unset TNS_ADMIN; \
                                         unset TWO_TASK; \
                                         export LD_LIBRARY_PATH="#{home}/lib"; \
                                         #{command}'`
        end

        def get_product_version(oracle_home)
          product_version = { 'name'    => get_home_name(oracle_home),
                              'os_user' => user_for_file(oracle_home),
                              'version' => get_version(oracle_home)
                            }
          cleanup_opatch_residue(oracle_home) if Facter.value(:kernel) == 'Linux'
          product_version
        end
      end
    end
  end
end

